/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.yuandian.dataflow.statemachine;

import org.apache.ratis.client.RaftClient;
import org.apache.ratis.conf.Parameters;
import org.apache.ratis.conf.RaftProperties;
// import org.apache.ratis.examples.common.Constants; 
import org.apache.ratis.grpc.GrpcConfigKeys;
import org.apache.ratis.grpc.GrpcFactory;
import org.apache.ratis.protocol.ClientId;
import org.apache.ratis.protocol.RaftGroup;
import org.apache.ratis.protocol.RaftGroupId;
import org.apache.ratis.protocol.RaftPeer;
import org.apache.ratis.server.RaftServer;
import org.apache.ratis.server.RaftServerConfigKeys;
import org.apache.ratis.util.NetUtils;

import com.yuandian.dataflow.statemachine.grpc.ProcessorServer;
import com.yuandian.dataflow.statemachine.state.Peer;
import com.yuandian.dataflow.statemachine_old.MasterFactory;

import io.netty.util.internal.StringUtil;
import lombok.Getter;
import lombok.Setter;
import lombok.extern.slf4j.Slf4j;

import java.io.Closeable;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Scanner;
import java.util.UUID;

import static java.nio.charset.StandardCharsets.UTF_8;

/**
 * Simplest Ratis server, use a simple state machine {@link StateMachine}
 * which maintain a counter across multi server.
 * This server application designed to run several times with different
 * parameters (1,2 or 3). server addresses hard coded in {@link Constants}
 * <p>
 * Run this application three times with three different parameter set-up a
 * ratis cluster which maintain a counter value replicated in each server memory
 */
@Slf4j
@Getter
@Setter
public final class StateServer implements Closeable {

  public static  HashMap<Peer, Peer> activesPeers = new HashMap<>();

  private   RaftClient raftClient  ;
  private final RaftServer raftServer;
  private final  RaftGroup raftGroupConf;
  private final ProcessorServer processorServer;

  private  Peer peer = new Peer() ;
 

  public static final UUID CLUSTER_GROUP_ID = UUID.fromString("79642d72-6166-742d-6461-7461666c6f77");

  public StateServer(RaftPeer curpeer,  ArrayList<RaftPeer> peers) throws IOException   {

    final File storageDir = new File("./raftdata/" + curpeer.getId());

    //create a property object
    RaftProperties properties = new RaftProperties();

    //set the storage directory (different for each peer) in RaftProperty object
    RaftServerConfigKeys.setStorageDir(properties, Collections.singletonList(storageDir));

    //set the port which server listen to in RaftProperty object
    final int port = NetUtils.createSocketAddr(curpeer.getAddress()).getPort();
    GrpcConfigKeys.Server.setPort(properties, port);
   
    //create the counter state machine which hold the counter value
    StateMachine stateMachine = new StateMachine();
  
    raftGroupConf = RaftGroup.valueOf(
      RaftGroupId.valueOf(CLUSTER_GROUP_ID), peers);
    
 
    //create and start the Raft server
    this.raftServer = RaftServer.newBuilder()
        .setGroup(raftGroupConf)
        .setProperties(properties)
        .setServerId(curpeer.getId())
        .setStateMachine(stateMachine)
        .build();
 
        raftClient = buildClient(raftGroupConf);
          // create RaftClient
   
      this.processorServer = new ProcessorServer();
      this.processorServer.getGrpcServer().start();

      this.peer.setProcessorPort(this.processorServer.getGrpcServer().getPort());
  }

  // block
  public void start() throws IOException, InterruptedException {
    raftServer.start();
    this.processorServer.getGrpcServer().awaitTermination();
  }

  @Override
  public void close() throws IOException {
    MasterFactory.getMasterExecute().interrupt();
    raftServer.close();
    this.processorServer.getGrpcServer().shutdown();
    
  }

  private static RaftClient buildClient( RaftGroup raftGroup) {
    RaftProperties raftProperties = new RaftProperties();
 
    RaftClient.Builder builder = RaftClient.newBuilder()
        .setProperties(raftProperties)
        .setRaftGroup(raftGroup)
        .setClientRpc(new GrpcFactory(new Parameters()).newRaftClientRpc(ClientId.randomId(), raftProperties));
    return builder.build();
  }

  public static void main(String[] args) throws IOException, InterruptedException {
    if (args.length < 1) {
      System.err.println("Usage: java -cp *.jar org.apache.ratis.examples.counter.server.StateServer {serverIndex}");
      System.err.println("{serverIndex} could be 1, 2 or 3");
      System.exit(1);
    }

    StateServerFactory.startServer(args[0]);

    // var peers = new ArrayList<RaftPeer>();
    // String[] addresses = new String[]{"localhost:4440","localhost:4441","localhost:4442"};

    // for (int i = 0; i < addresses.length; i++) {
    //   // var port = addresses[i].split(":")[1]; 
    //   peers.add(RaftPeer.newBuilder().setId("yd-" + args[0]).setAddress(addresses[i]).build());
    // }

    // //find current peer object based on application parameter
    // final RaftPeer currentPeer = peers.get(Integer.parseInt(args[0]));

    // //start a counter server
    // final File storageDir = new File("./raftdata/" + currentPeer.getId());
    // final StateServer stateServer = new StateServer(currentPeer, peers, storageDir);

  

    // stateServer.start();
    // stateServer.close();
  }
}
